;;
;; BriefLZ  -  small fast Lempel-Ziv
;;
;; MASM assembler include file
;;
;; Copyright (c) 2002-2005 by Joergen Ibsen / Jibz
;; All Rights Reserved
;;
;; http://www.ibsensoftware.com/
;;
;; This software is provided 'as-is', without any express
;; or implied warranty.  In no event will the authors be
;; held liable for any damages arising from the use of
;; this software.
;;
;; Permission is granted to anyone to use this software
;; for any purpose, including commercial applications,
;; and to alter it and redistribute it freely, subject to
;; the following restrictions:
;;
;; 1. The origin of this software must not be
;;    misrepresented; you must not claim that you
;;    wrote the original software. If you use this
;;    software in a product, an acknowledgment in
;;    the product documentation would be appreciated
;;    but is not required.
;;
;; 2. Altered source versions must be plainly marked
;;    as such, and must not be misrepresented as
;;    being the original software.
;;
;; 3. This notice may not be removed or altered from
;;    any source distribution.
;;

IFNDEF BRIEFLZ_INC_INCLUDED
BRIEFLZ_INC_INCLUDED = 1

BLZ_ERROR = -1

; Compress data.
; @param source - pointer to data.
; @param destination - where to store the compressed data.
; @param srclength - the length in bytes of the data.
; @param workmem - pointer to memory for temporary use.
; @return the length of the compressed data.
;
blz_pack PROTO C source:DWORD,destination:DWORD,
                 srclength:DWORD,workmem:DWORD


; Decompress data.
; @param source - pointer to the compressed data.
; @param destination - where to store the decompressed data.
; @param depacked_length - the length of the decompressed data.
; @return the length of the decompressed data.
;
blz_depack PROTO C source:DWORD,destination:DWORD,
                   depacked_length:DWORD


; Decompress data safely.
; @param source - pointer to the compressed data.
; @param srclen - the size of the source buffer in bytes.
; @param destination - where to store the decompressed data.
; @param depacked_length - the length of the decompressed data.
; @return the length of the decompressed data, or BLZ_ERROR on error.
; @note This functions reads at most srclen bytes from source[], and
; writes at most depacked_length bytes to destination[].
;
blz_depack_safe PROTO C source:DWORD,srclen:DWORD,
                        destination:DWORD,depacked_length:DWORD


; Get the required size of the workmem buffer.
; @param srclength - the length in bytes of the data.
; @return required size in bytes of the workmem buffer.
;
blz_workmem_size PROTO C srclength:DWORD


; Get the maximum output size produced on uncompressible data.
; @param srclength - the length in bytes of the data.
; @return maximum required size in bytes of the output buffer.
;
blz_max_packed_size PROTO C srclength:DWORD


; Compute the CRC32 of a buffer.
; @param source - pointer to the data.
; @param srclength - the number of bytes to process.
; @param initial_crc32 - the current CRC32 value (pass 0 for first block).
; @return the CRC32 of the data.
;
blz_crc32 PROTO C source:DWORD,srclength:DWORD,
                  initial_crc32:DWORD

ENDIF
