;;
;; BriefLZ  -  small fast Lempel-Ziv
;;
;; MASM assembler packer
;;
;; Copyright (c) 2002-2005 by Joergen Ibsen / Jibz
;; All Rights Reserved
;;
;; http://www.ibsensoftware.com/
;;
;; This software is provided 'as-is', without any express
;; or implied warranty.  In no event will the authors be
;; held liable for any damages arising from the use of
;; this software.
;;
;; Permission is granted to anyone to use this software
;; for any purpose, including commercial applications,
;; and to alter it and redistribute it freely, subject to
;; the following restrictions:
;;
;; 1. The origin of this software must not be
;;    misrepresented; you must not claim that you
;;    wrote the original software. If you use this
;;    software in a product, an acknowledgment in
;;    the product documentation would be appreciated
;;    but is not required.
;;
;; 2. Altered source versions must be plainly marked
;;    as such, and must not be misrepresented as
;;    being the original software.
;;
;; 3. This notice may not be removed or altered from
;;    any source distribution.
;;

; must be a power of 2 (between 8k and 2mb appear reasonable)
BLZ_WORKMEM_SIZE EQU 1024*1024

IF BLZ_WORKMEM_SIZE AND (BLZ_WORKMEM_SIZE-1)
  .ERR <BLZ_WORKMEM_SIZE must be a power of 2>
ENDIF

.386
.MODEL FLAT,C

OPTION CASEMAP:NONE

PUBLIC blz_workmem_size
PUBLIC blz_max_packed_size
PUBLIC blz_pack

.CODE

; =============================================================

putbit0M MACRO                ; add a 0-bit to the current tag word
  LOCAL bitsleft, done
    test   ebp, ebp
    jnz    bitsleft
    mov    edx, edi
    inc    ebp
    add    edi, 2
  bitsleft:
    add    bp, bp
    jnc    done
    mov    [edx], bp
    xor    ebp, ebp
  done:
ENDM putbit0M

putbit1M MACRO                ; add a 1-bit to the current tag word
  LOCAL bitsleft, done
    test   ebp, ebp
    jnz    bitsleft
    mov    edx, edi
    inc    ebp
    add    edi, 2
  bitsleft:
    add    bp, bp
    inc    bp
    jnc    done
    mov    [edx], bp
    xor    ebp, ebp
  done:
ENDM putbit1M

putbitM MACRO                 ; add bit according to carry
  LOCAL onebit, bitdone
    jc     onebit
    putbit0M
    jmp    bitdone
  onebit:
    putbit1M
  bitdone:
ENDM putbitM

putgammaM MACRO               ; output gamma coding of value in eax (eax > 1)
  LOCAL revmore, outmore, outstart
    push   ebx
    push   eax
    shr    eax, 1
    mov    ebx, 1
  revmore:
    shr    eax, 1
    jz     outstart
    adc    ebx, ebx
    jmp    revmore
  outmore:
    putbitM
    putbit1M
  outstart:
    shr    ebx, 1
    jnz    outmore
    pop    eax
    shr    eax, 1
    putbitM
    putbit0M
    pop    ebx
ENDM putgammaM

hash4M MACRO                  ; hash next 4 bytes into eax
    push   ecx
    movzx  eax, byte ptr [esi]
    imul   eax, 317              ; the imuls can be replaced with:
    movzx  ecx, byte ptr [esi+1] ;
    add    eax, ecx              ;   lea    ecx, [eax*2 + eax]
    imul   eax, 317              ;   shl    eax, 6
    movzx  ecx, byte ptr [esi+2] ;   lea    eax, [eax*4 + eax]
    add    eax, ecx              ;   sub    eax, ecx
    imul   eax, 317
    movzx  ecx, byte ptr [esi+3]
    add    eax, ecx
    and    eax, (BLZ_WORKMEM_SIZE/4)-1
    pop    ecx
ENDM hash4M

; =============================================================

blz_workmem_size:
    ; blz_workmem_size(unsigned int length);

    mov    eax, BLZ_WORKMEM_SIZE

    ret

; =============================================================

blz_max_packed_size:
    ; blz_max_packed_size(unsigned int length);

    _slen$ EQU 4

    mov    eax, [esp + _slen$] ; length + length/8 + 64
    mov    edx, eax
    shr    edx, 3
    lea    eax, [eax + edx + 64]

    ret

; =============================================================

blz_pack:
    ; blz_pack(const void *source,
    ;          void *destination,
    ;          unsigned int length,
    ;          void *workmem);

    _wkm$  EQU 2*4 + 4*4 + 4 + 12
    _len$  EQU 2*4 + 4*4 + 4 + 8
    _dst$  EQU 2*4 + 4*4 + 4 + 4
    _src$  EQU 2*4 + 4*4 + 4

    _lim$  EQU 4
    _bpt$  EQU 0

    push   ebx
    push   esi
    push   edi
    push   ebp

    sub    esp, 2*4           ; make room for temps

    cld

    mov    edi, [esp + _wkm$] ; edi -> lookuptable[]

    xor    eax, eax           ; clear lookuptable[]
    mov    ecx, BLZ_WORKMEM_SIZE/4
    rep    stosd

    mov    esi, [esp + _src$]
    mov    edi, [esp + _dst$]
    mov    ebx, [esp + _len$]

    lea    eax, [ebx + esi - 4]
    mov    [esp + _lim$], eax ; limit = source + length - 4

    mov    [esp + _bpt$], esi ; backptr = source

    test   ebx, ebx           ; length == 0?
    jz     EODdone            ;

    mov    al, [esi]          ; fist byte verbatim
    inc    esi                ;
    mov    [edi], al          ;
    inc    edi                ;

    cmp    ebx, 1             ; only one byte?
    je     EODdone            ;

    mov    bp, 1              ; initialise tag
    mov    edx, edi           ;
    add    edi, 2             ;

    jmp    nexttagcheck

  no_match:
    putbit0M                  ; 0-bit = literal

    mov    al, [esi]          ; copy literal
    inc    esi                ;
    mov    [edi], al          ;
    inc    edi                ;

  nexttagcheck:
    cmp    esi, [esp + _lim$] ; are we done?
    jae    donepacking        ;

  nexttag:
    mov    ecx, [esp + _wkm$] ; ecx -> lookuptable[]

    mov    ebx, esi           ; ebx = buffer - backptr
    mov    esi, [esp + _bpt$] ; i.e. distance from backptr to current
    sub    ebx, esi           ;

  update:
    hash4M                    ; hash next 4 bytes

    mov    [ecx + eax*4], esi ; lookuptable[hash] = backptr

    inc    esi                ; ++backptr
    dec    ebx
    jnz    update

    mov    [esp + _bpt$], esi ; esi is now back to current pos

    hash4M                    ; hash next 4 bytes

    mov    ebx, [ecx + eax*4] ; ebx = lookuptable[hash]

    test   ebx, ebx           ; no match?
    jz     no_match           ;

    ; check match length
    mov    ecx, [esp + _lim$] ; ecx = max allowed match length
    sub    ecx, esi           ;
    add    ecx, 4             ;

    push   edx

    xor    eax, eax
  compare:
    mov    dl, [ebx + eax]    ; compare possible match with current
    cmp    dl, [esi + eax]    ;
    jne    matchlen_found

    inc    eax

    dec    ecx
    jnz    compare

  matchlen_found:
    pop    edx

    cmp    eax, 4             ; match too short?
    jb     no_match           ;

    mov    ecx, esi           ;
    sub    ecx, ebx           ; ecx = matchpos

    putbit1M                  ; 1-bit = match

    add    esi, eax           ; update esi to next position

    sub    eax, 2             ; matchlen >= 4, so subtract 2

    putgammaM                 ; output gamma coding of matchlen - 2

    dec    ecx                ; matchpos > 0, so subtract 1

    mov    eax, ecx           ; eax = (matchpos >> 8) + 2
    shr    eax, 8             ;
    add    eax, 2             ;

    putgammaM                 ; output gamma coding of (matchpos >> 8) + 2

    mov    [edi], cl          ; output low 8 bits of matchpos
    inc    edi                ;

    cmp    esi, [esp + _lim$] ; are we done?
    jb     nexttag            ;

  donepacking:
    mov    ebx, [esp + _lim$] ; ebx = source + length
    add    ebx, 4             ;

    jmp    check_final_literals

  final_literals:
    putbit0M                  ; 0-bit = literal

    mov    al, [esi]          ; copy literal
    inc    esi                ;
    mov    [edi], al          ;
    inc    edi                ;

  check_final_literals:
    cmp    esi, ebx
    jb     final_literals

    test   ebp, ebp           ; do we need to fix the last tag?
    jz     EODdone            ;

  doEOD:
    add    bp, bp             ; shift last tag into position
    jnc    doEOD              ;

    mov    [edx], bp          ; and put it into it's place

  EODdone:
    mov    eax, edi           ; return packed length in eax
    sub    eax, [esp + _dst$] ;

    add    esp, 2*4

    pop    ebp
    pop    edi
    pop    esi
    pop    ebx

    ret

; =============================================================

END

; =============================================================
