;;
;; BriefLZ  -  small fast Lempel-Ziv
;;
;; MASM safe assembler depacker
;;
;; Copyright (c) 2002-2005 by Joergen Ibsen / Jibz
;; All Rights Reserved
;;
;; http://www.ibsensoftware.com/
;;
;; This software is provided 'as-is', without any express
;; or implied warranty.  In no event will the authors be
;; held liable for any damages arising from the use of
;; this software.
;;
;; Permission is granted to anyone to use this software
;; for any purpose, including commercial applications,
;; and to alter it and redistribute it freely, subject to
;; the following restrictions:
;;
;; 1. The origin of this software must not be
;;    misrepresented; you must not claim that you
;;    wrote the original software. If you use this
;;    software in a product, an acknowledgment in
;;    the product documentation would be appreciated
;;    but is not required.
;;
;; 2. Altered source versions must be plainly marked
;;    as such, and must not be misrepresented as
;;    being the original software.
;;
;; 3. This notice may not be removed or altered from
;;    any source distribution.
;;

.386
.MODEL FLAT,C

OPTION CASEMAP:NONE

PUBLIC blz_depack_safe

.CODE

; =============================================================

getbitM MACRO                 ; get next tag-bit into carry
  LOCAL stillbitsleft
    add    dx, dx
    jnz    stillbitsleft

    sub    ebp, 2             ; read two bytes from source
    jc     return_error       ;

    mov    dx, [esi]
    add    esi, 2

    add    dx, dx
    inc    dx
  stillbitsleft:
ENDM getbitM

domatchM MACRO reg            ; copy a match, ecx = len, param = pos
    push   ecx
    mov    ecx, [esp + 4 + _dlen$] ; ecx = dstlen
    sub    ecx, ebx                ; ecx = num written
    cmp    reg, ecx
    pop    ecx
    ja     return_error

    sub    ebx, ecx           ; write ecx bytes to destination
    jc     return_error       ;

    push   esi
    mov    esi, edi
    sub    esi, reg
    rep    movsb
    pop    esi
ENDM domatchM

getgammaM MACRO reg           ; gamma decode value into param
  LOCAL getmore
    mov    reg, 1
  getmore:
    getbitM
    adc    reg, reg
    jc     return_error
    getbitM
    jc     getmore
ENDM getgammaM

; =============================================================

blz_depack_safe:
    ; blz_depack_safe(const void *source,
    ;                 unsigned int srclen,
    ;                 void *destination,
    ;                 unsigned int dstlen);

    _dlen$ EQU 4*4 + 4 + 12
    _dst$  EQU 4*4 + 4 + 8
    _slen$ EQU 4*4 + 4 + 4
    _src$  EQU 4*4 + 4

    push   ebx
    push   ebp
    push   esi
    push   edi

    mov    esi, [esp + _src$]
    mov    ebp, [esp + _slen$]
    mov    edi, [esp + _dst$]
    mov    ebx, [esp + _dlen$]

    cld
    mov    dx, 8000h          ; initialise tag

  literal:
    sub    ebp, 1             ; read one byte from source
    jc     return_error       ;

    mov    al, [esi]          ; read literal
    inc    esi                ;

    sub    ebx, 1             ; write one byte to destination
    jc     return_error       ;

    mov    [edi], al          ; write literal
    inc    edi                ;

    test   ebx, ebx           ; are we done?
    jz     donedepacking

  nexttag:
    getbitM                   ; literal or match?
    jnc    literal            ;

    getgammaM ecx             ; ecx = matchlen

    getgammaM eax             ; eax = (matchpos >> 8) + 2

    add    ecx, 2             ; matchlen >= 4, so add 2

    shl    eax, 8             ; eax = high part of matchpos

    sub    ebp, 1             ; read one byte from source
    jc     return_error       ;

    mov    al, [esi]          ; add low 8 bits of matchpos
    inc    esi                ;

    add    eax, 0fffffe01h    ; adjust matchpos (highpart-=2, lowpart+=1)

    domatchM eax              ; copy match

    test   ebx, ebx           ; are we done?
    jnz    nexttag

  donedepacking:
    mov    eax, edi           ; return unpacked length in eax
    sub    eax, [esp + _dst$] ;

    jmp    return_eax

  return_error:
    or     eax, -1

  return_eax:
    pop    edi
    pop    esi
    pop    ebp
    pop    ebx

    ret

; =============================================================

END

; =============================================================
