;;
;; BriefLZ  -  small fast Lempel-Ziv
;;
;; MASM small safe assembler depacker
;;
;; Copyright (c) 2002-2005 by Joergen Ibsen / Jibz
;; All Rights Reserved
;;
;; http://www.ibsensoftware.com/
;;
;; This software is provided 'as-is', without any express
;; or implied warranty.  In no event will the authors be
;; held liable for any damages arising from the use of
;; this software.
;;
;; Permission is granted to anyone to use this software
;; for any purpose, including commercial applications,
;; and to alter it and redistribute it freely, subject to
;; the following restrictions:
;;
;; 1. The origin of this software must not be
;;    misrepresented; you must not claim that you
;;    wrote the original software. If you use this
;;    software in a product, an acknowledgment in
;;    the product documentation would be appreciated
;;    but is not required.
;;
;; 2. Altered source versions must be plainly marked
;;    as such, and must not be misrepresented as
;;    being the original software.
;;
;; 3. This notice may not be removed or altered from
;;    any source distribution.
;;

.386
.MODEL FLAT,C

OPTION CASEMAP:NONE

PUBLIC blz_depack_safe

.CODE

; =============================================================

blz_depack_safe:
    ; blz_depack_safe(const void *source,
    ;                 unsigned int srclen,
    ;                 void *destination,
    ;                 unsigned int dstlen);

    _dlen$ EQU 4*4 + 4 + 12
    _dst$  EQU 4*4 + 4 + 8
    _slen$ EQU 4*4 + 4 + 4
    _src$  EQU 4*4 + 4

    push   ebx
    push   ebp
    push   esi
    push   edi

    mov    esi, [esp + _src$]
    mov    ebp, [esp + _slen$]
    mov    edi, [esp + _dst$]
    mov    ebx, [esp + _dlen$]

    cld
    xor    edx, edx           ; initialise tag

    push   edx                ; push sentinel

  literal:
    dec    ebp                ; read one byte from source
    jc     return_error       ;

    dec    ebx                ; write one byte to destination
    jc     return_error       ;

    movsb                     ; copy literal

  nexttag:
    test   ebx, ebx           ; are we done?
    jz     donedepacking

    call   getbit             ; literal or match?
    jnc    literal            ;

    call   getgamma           ; ecx = matchlen
    xchg   eax, ecx           ;

    call   getgamma           ; eax = (matchpos >> 8) + 2

    dec    eax                ; eax = (matchpos >> 8)
    dec    eax                ;

    inc    ecx                ; matchlen >= 4, so add 2
    inc    ecx                ;

    dec    ebp                ; read one byte from source
    jc     return_error       ;

    shl    eax, 8             ; eax = high part of matchpos
    lodsb                     ; add low 8 bits of matchpos

    inc    eax                ; matchpos > 0, so add 1

    push   ecx
    mov    ecx, [esp + 8 + _dlen$] ; ecx = dstlen
    sub    ecx, ebx                ; ecx = num written
    cmp    eax, ecx
    pop    ecx
    ja     return_error

    sub    ebx, ecx           ; write ecx bytes to destination
    jc     return_error       ;

    push   esi

    mov    esi, edi           ; copy match
    sub    esi, eax           ;
    rep    movsb              ;

    pop    esi

    jmp    nexttag

  getbit:                     ; get next tag-bit into carry
    add    dx, dx
    jnz    stillbitsleft

    sub    ebp, 2             ; read two bytes from source
    jc     return_error       ;

    xchg   eax, edx
    lodsw
    xchg   eax, edx
    add    dx, dx
    inc    edx
  stillbitsleft:
    ret

  getgamma:                   ; gamma decode value into eax
    xor    eax, eax
    inc    eax
  getmore:
    call   getbit
    adc    eax, eax
    jc     return_error
    call   getbit
    jc     getmore
    ret

  return_error:
    or     eax, -1

    jmp    find_sentinel

  donedepacking:
    xchg   eax, edi               ; return unpacked length in eax
    sub    eax, [esp + 4 + _dst$] ;

  find_sentinel:
    pop    ecx
    test   ecx, ecx
    jnz    find_sentinel

  return_eax:
    pop    edi
    pop    esi
    pop    ebp
    pop    ebx

    ret

; =============================================================

END

; =============================================================
